<?php

/**
 * Implementation of hook_ext_model_types.
 */
function ext_ext_model_types() {
  $types = array();
  foreach (array('node', 'user', 'view') as $type) {
    $types[$type] = array(
      'user_created' => FALSE,
    );
  }
  return $types;
}

/**
 * Implementation of hook_ext_store_types.
 */
function ext_ext_store_types() {
  $types = array();
  foreach (array('node', 'user', 'file', 'comment', 'view') as $type) {
    $types[$type] = array(
      'user_created' => FALSE,
    );
  }
  return $types;
}

/**
 * Implementation of hook_ext_models.
 */
function ext_ext_models() {
  $models = array();

  // TODO: Create Model defs for terms.

  // Create Model for node.
  $type = array(
    '#type' => 'node',
    '#db_base_table' => 'node',
    'fields' => array(),
    'idProperty' => 'nid',
  );
  // Get standard fields for nodes.
  $schema = drupal_get_schema('node');
  foreach ($schema['fields'] as $field_name => $field) {
    $type['fields'][] = _ext_db_schema_field_to_ext_field($field_name, $field, 'node');
  }
  // Remove unused field(s).
  unset($type['fields']['moderate']);
  // Add some fields from node_revisions.
  $schema = drupal_get_schema('node_revisions');
  foreach (array('body', 'teaser') as $field_name) {
    $type['fields'][] = _ext_db_schema_field_to_ext_field($field_name, $schema[$field_name], 'node_revisions');
  }

  // Add standard associations.
  $type['associations'] = array(
    array(
      'type' => 'belongsTo',
      'model' => 'User',
      'primaryKey' => 'uid',
      'foreignKey' => 'uid',
    ),
  );

  $models['Node'] = $type;


  // Create model for each content type.
  if (module_exists('content')) {
    // Get info about all CCK fields for all content types.
    $cck_fields = _content_type_info();
  }
  foreach (node_get_types() as $node_type => $node_type_info) {
    $type = array(
      '#type' => 'node',
      '#node_type' => $node_type,
      '#db_base_table' => 'node',
      '#extends' => 'Node',
      'idProperty' => 'nid',
      'fields' => array(),
      'associations' => array(),
      'validations' => array(),
    );

    // Add standard validations.
    $type['validations'] = array();
    if ($node_type_info->has_title) {
      $type['validations'][] = array(
        'field' => 'title',
        'type' => 'presence',
      );
    }

    // Add CCK fields if available.
    if (module_exists('content')) {
      foreach ($cck_fields['content types'][$node_type]['fields'] as $field_name => $field) {
        if (isset($field['widget']['default_value'][0]) && is_array($field['widget']['default_value'][0])) {
          $default_value = $field['widget']['default_value'][0];
        }
        else {
          $default_value = FALSE;
        }

        $db_info = content_database_info($field);

        // Add field for each (DB) column the CCK field defines.
        $columns = array_keys($field['columns']);
        foreach ($columns as $col) {
          $field_def = array(
            'name' => $db_info['columns'][$col]['column'], // Should be $field_name. '_'. $col
            '#type' => _ext_db_schema_field_type_to_ext_field_type($db_info['columns'][$col]['type']),
            '#db_table' => $db_info['table'],
            '#db_column' => $db_info['columns'][$col]['column'],
            // Store some info about the CCK field to make things easier later.
            '#content_field_name' => $field_name, // The CCK field name.
            '#content_column' => $col, // The value key in the field on the node.
          );

          // Only specify a type for fields that have a single value and are not
          // serialized, otherwise Ext tries to convert an array of values, or some
          // other structure to a single value.
          if (!$field['multiple'] && !$db_info['columns'][$col]['serialize']) {
            $field_def['type'] = $field_def['#type'];
          }

          if ($default_value) {
            $field_def['defaultValue'] = $default_value[$col];
          }

          $type['fields'][] = $field_def;
        }

        // Special handling for filefield type to include extra columns from 'files' table.
        if ($field['type'] == 'filefield') {
          $schema = drupal_get_schema('files');
          foreach ($schema['fields'] as $files_field_name => $files_field) {
            // 'fid' included in CCK filefield def.
            if ($files_field_name != 'fid') {
              $field_def = array(
                'name' => $field_name. '_'. $files_field_name,
                '#type' => _ext_db_schema_field_type_to_ext_field_type($files_field['type']),
                '#db_table' => 'files',
                '#db_column' => $files_field_name,
                '#content_field_name' => $field_name, // The CCK field name.
                '#content_column' => $files_field_name, // The value key in the field on the node.
              );
              if (isset($field['default'])) {
                $field_def['defaultValue'] = $files_field['default'];
              }
              // Only specify a type for fields that have a single value and are not
              // serialized, otherwise Ext tries to convert an array of values, or some
              // other structure to a single value.
              if (!$field['multiple'] && !$files_field['serialize']) {
                $field_def['type'] = $field_def['#type'];
              }
              $type['fields'][] = $field_def;
            }
          }
        }

        // First field name, used in validations and/or associations.
        $field_col_name = $field_name. '_'. $columns[0];

        // If there is only one column containing a single value, add standard
        // validations.
        // TODO provide validations for multiple column field types and/or multiple value fields?
        if (count($columns) == 1  && !$field['multiple']) {
          if ($field['required']) {
            $type['validations'][] = array(
              'field' => $field_col_name,
              'type' => 'presence',
            );
          }
          if (!empty($field['max_length'])) {
            $type['validations'][] = array(
              'field' => $field_col_name,
              'type' => 'length',
              'max' => $field['max_length'],
            );
          }
          // The "allowed values" validation is determined at run time when it
          // is calculated using a PHP script.
          if (!empty($field['allowed_values']) && empty($field['allowed_values_php'])) {
            $type['validations'][] = array(
              'field' => $field_col_name,
              'type' => 'inclusion',
              'list' => array_keys(content_allowed_values($field)),
            );
          }
          elseif (!empty($field['allowed_values_php'])) {
            $type['server_validate'] = TRUE;
          }
        }

        // Add associations if available.
        // We only support single value CCK fields (mapped as 'belongsTo' association types)
        // at the moment. Conceptually, multiple value fields could be mapped as a 'hasMany'
        // association, however the foreignKey for a hasMany association is the field name
        // on the associated Model linking it to the owner Model, and there is no such field
        // on referenced nodes (one could be added under certain circumstances (ie where it's
        // guaranteed that only one reference will exist to a node from this particular CCK
        // field type), but this is beyond the scope of this module.
        if (($field['type'] == 'nodereference' || $field['type'] == 'userreference') && !$field['multiple']) {
          if ($field['type'] == 'nodereference') {
            // Create association for every node type that can be referenced.
            foreach (array_filter($field['referenceable_types']) as $ref_type) {
              $ref_model_name = 'Node'. _ext_camelcase($ref_type);
              $assoc = array(
                'type' => 'belongsTo',
                'model' => $ref_model_name,
                'primaryKey' => 'nid', // Primary ID of associated Model.
                'foreignKey' => $field_col_name, // Field name on this Model linking it to associated Model.
                //'#load_nested_max_depth' => 1,
              );
              $type['associations'][] = $assoc;
            }
          }
          else { // $field['type'] == 'userreference'
            $assoc = array(
              'type' => 'belongsTo',
              'model' => 'User',
              'primaryKey' => 'uid', // Primary ID of associated Model.
              'foreignKey' => $field_col_name, // Field name on this Model linking it to associated Model.
            );
            $type['associations'][] = $assoc;
          }
        }
      }
    }

    $models[_ext_gen_name('node', $node_type)] = $type;
  }


  // Create Model for user.
  $type = array(
    '#type' => 'user',
    '#db_base_table' => 'users',
    'fields' => array(),
    'idProperty' => 'uid',
  );
  // Get standard fields for users.
  $schema = drupal_get_schema('users');
  foreach ($schema['fields'] as $field_name => $field) {
    if (!in_array($field_name, array('pass', 'mode', 'sort', 'threshold', 'theme', 'signature_format', 'data'))) {
      $type['fields'][] = _ext_db_schema_field_to_ext_field($field_name, $field, 'users');
    }
  }
  // Add standard validations (below regexs for email validation).
  $user = '[a-zA-Z0-9_\-\.\+\^!#\$%&*+\/\=\?\`\|\{\}~\']+';
  $domain = '(?:(?:[a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.?)+';
  $ipv4 = '[0-9]{1,3}(\.[0-9]{1,3}){3}';
  $ipv6 = '[0-9a-fA-F]{1,4}(\:[0-9a-fA-F]{1,4}){7}';
  $type['validations'] = array(
    array(
      'field' => 'name',
      'type' => 'presence',
    ),
    array(
      'field' => 'mail',
      'type' => 'format',
      'matcher' => "^$user@($domain|(\[($ipv4|$ipv6)\]))\$",
    ),
  );
  if (module_exists('profile')) {
    // Add profile fields.
    foreach (profile_categories() as $cat_info) {
      $profile_fields = _profile_get_fields($cat_info['name']);
      while ($field = db_fetch_array($profile_fields)) {
        $field_name = $field['name'];
        // Add field.
        $type['fields'][] = array(
          'name' => $field_name,
          'type' => 'string', // All profile field values are stored as text.
        );
        // Add validations.
        $type['validations'] = array();
        if ($field['required']) {
          $type['validations'][] = array(
            'field' => $field_name,
            'type' => 'presence',
          );
        }
        if ($field['type'] == 'selection') {
          $type['validations'][] = array(
            'field' => $field_name,
            'type' => 'inclusion',
            'list' => array_filter(split("[\n\r]", $field['options'])),
          );
        }
      }
    }
  }
  $models['User'] = $type;


  // Create Model for files.
  $type = array(
    '#type' => 'file',
    '#db_base_table' => 'files',
    'idProperty' => 'fid',
    'fields' => array(), // Added below from schema.
    'associations' => array(
      array(
        'type' => 'belongsTo',
        'model' => 'User',
        'primaryKey' => 'uid',
        'foreignKey' => 'uid',
      ),
    ),
  );
  // Get standard fields for files.
  $schema = drupal_get_schema('files');
  foreach ($schema['fields'] as $field_name => $field) {
    $type['fields'][] = _ext_db_schema_field_to_ext_field($field_name, $field, 'files');
  }
  $models['File'] = $type;


  if (module_exists('comment')) {
    // Create Model for comments.
    $type = array(
      '#type' => 'comment',
      '#db_base_table' => 'comments',
      'idProperty' => 'cid',
      'fields' => array(), // Added below from schema.
      'associations' => array(
        array(
          'type' => 'belongsTo',
          'model' => 'User',
          'primaryKey' => 'uid',
          'foreignKey' => 'uid',
        ),
        array(
          'type' => 'belongsTo',
          'model' => 'Comment',
          'primaryKey' => 'uid',
          'foreignKey' => 'pid',
        ),
        array(
          'type' => 'belongsTo',
          'model' => 'Node',
          'primaryKey' => 'nid',
          'foreignKey' => 'nid',
        ),
      ),
    );
    // Get standard fields for comments.
    $schema = drupal_get_schema('comments');
    foreach ($schema['fields'] as $field_name => $field) {
      $type['fields'][] = _ext_db_schema_field_to_ext_field($field_name, $field, 'files');
    }
    $models['Comment'] = $type;
  }


  if (module_exists('views')) {
    // Create Models for relevant Views displays.
    // These are used for View displays that use a 'fields' row style.
    $views = views_get_all_views();
    foreach ($views as $view_name => $view) {
      if (!$view->disabled) {
        $view->set_display(); // Builds all settings for View displays.

        // Create a Model for each display that has row style set to 'fields'.
        foreach ($view->display as $display_name => $display) {
          if ($display->handler->options['row_plugin'] == 'fields' && isset($display->display_options['fields'])) {
            $type_name = _ext_gen_name('view', $view_name, $display_name);
            $type = array(
              '#type' => 'view',
              '#view_name' => $view_name,
              '#display_name' => $display_name,
              'idProperty' => _ext_table_primary_id($view->base_table),
              'fields' => array(),
            );
            foreach ($display->display_options['fields'] as $field_name => $field) {
              $type['fields'][] = array(
                'name' => $field['table']. '_'. $field['field'],
              );
            }
            $models[_ext_gen_name('view', $view_name, $display_name)] = $type;
          }
        }
      }
    }
  }

  return $models;
}


/**
 * Implementation of hook_ext_stores.
 */
function ext_ext_stores() {
  $stores = array();

  // Create generic node and user Store definitions, one for each
  // node type, and the user, file and comment models.
  $models = ext_ext_models();
  foreach ($models as $model_name => $model) {
    // If the Model is for a node type, file, comment or user.
    if (in_array($model['#type'], array('node', 'user', 'file', 'comment'))) {
      $store = array(
        '#type' => $model['#type'],
        '#generic' => TRUE,
        'model' => $model_name,
        '#filterable_fields' => array(),
        '#sortable_fields' => array(),
      );

      // Collect filterable and sortable fields from base Node Model.
      if (isset($model['#extends']) && $model['#extends'] == 'Node') {
        $store['#filterable_fields'] = $stores['Node']['#filterable_fields'];
        $store['#sortable_fields'] = $stores['Node']['#sortable_fields'];
      }

      foreach ($model['fields'] as $field) {
        $store['#filterable_fields'][] = $field['name'];

        // If this is a CCK field then check if it's sortable.
        if (isset($field['#content_field_name'])) {
          $field = content_fields($field['#content_field_name'], $model['#node_type']);
          $db_info = content_database_info($field);
          if (!empty($db_info['columns'][$field['#content_column']]['sortable'])) {
            $store['#sortable_fields'][] = $field['name'];
          }
        }
        else {
          $store['#sortable_fields'][] = $field['name'];
        }
      }
      $stores[$model_name] = $store;
    }
  }

  if (module_exists('views')) {
    $views = views_get_all_views();
    foreach ($views as $view_name => $view) {
      if (!$view->disabled) {
        $view->set_display(); // Builds all options for View displays.
        // Create a Store for each display.
        foreach ($view->display as $display_name => $display) {
          $type_name = _ext_gen_name('view', $view_name, $display_name);
          $type = array(
            '#type' => 'view',
          ); // Store def.

          // If the 'fields' row style is not specified.
          if ($display->handler->options['row_plugin'] != 'fields') {
            // This View display can only be used as an instance selector, we use
            // one of the predefined Models.
            if ($view->base_table == 'node') {
              // Only Display types that filter the type of node to one specific
              // content type can be used.
              $filters = $display->display_options['filters'];
              if (isset($filters['type']) &&
                  $filters['type']['operator'] == 'in' &&
                  count($filters['type']['value']) == 1) {
                // Set the Model to that of the content type.
                $content_type = array_pop($filters['type']['value']);
                $type['model'] = _ext_gen_name('node', $content_type);
              }
            }
            else {
              $model_name = _ext_base_table_to_model_name($view->base_table);
              if ($model_name) {
                $type['model'] = $model_name;
              }
            }
          }
          // If the 'fields' row style is specified and fields have been specified.
          elseif (isset($display->display_options['fields'])) {
            // Use View/display specific Model.
            $type['model'] = $type_name;
            // Record the primary ID field for later use.
            $type['#primaryID'] = _ext_table_primary_id($view->base_table);
          }

          // Only add the Store def if we could determine a Model to use.
          if (isset($type['model'])) {
            $stores[$type_name] = $type;
          }
        }
      }
    }
  }

  return $stores;
}


/**
 * Implementation of hook_ext_access_model_data.
 */
function ext_ext_access_model_data($model_instance_id, $model_name, $op, $account) {
  $model = ext_get_models($model_name);
  if ($model['#type'] == 'node') {
    if ($op == 'create') {
      $node_type = substr(_ext_uncamelcase($model_name), 5);
      return node_access($op, $node_type, $account);
    }
    $node = node_load($model_instance_id);
    return node_access($op, $node, $account);
  }
  elseif ($model['#type'] == 'user') {
    if ($op == 'create')
      return user_access('administer users');
    $account = user_load($model_instance_id);
    if ($op == 'update' || $op == 'delete')
      return user_edit_access($account);
    if ($op == 'view')
      return user_view_access($account);
  }
  elseif ($model['#type'] == 'file') {
    // TODO: Creating and updating file model instances isn't supported at the moment, should they be?
    return $op == 'view';
  }
  elseif ($model['#type'] == 'comment') {
    if (user_access('access comments')) {
      if ($op == 'create') {
        // This isn't ideal, we also need to check that they can post the
        // comment to the given node. This is done in ext_ext_save_model_data.
        return user_access('post comments', $account);
      }
      else {
        $comment = _comment_load($model_instance_id);
        $node = node_load($comment->nid);
        // If they have edit access to the comment, they can do anything with it.
        // Logic borrowed from comment_access() function (modified to use arbitrary account).
        if (($account->uid && $account->uid == $comment->uid && comment_num_replies($comment->cid) == 0 && $comment->status == COMMENT_PUBLISHED) || user_access('administer comments', $account)) {
          return TRUE;
        }

        // For 'view', just check they have view access for the relevant node.
        if ($op == 'view' && node_access('view', $node, $account)) {
          return TRUE;
        }
      }
    }
  }
  return FALSE;
}


/**
 * Implementation of hook_ext_load_model_data.
 */
function ext_ext_load_model_data($model_instance_id, $model_name, $options=array()) {
  $model = ext_get_models($model_name);
  // 'view' data is only loaded via a Store so is not handled here.
  if ($model['#type'] == 'node') {
    $data = (array) node_load($model_instance_id);

    // Check node type against model name. In some cases this function can be
    // called for the wrong node type, eg when a CCK nodereference field can
    // reference multiple node types.
    if ($model_name != 'Node' && ($data['type'] != $model['#node_type']))
      return FALSE;

    _ext_convert_from_cck_fields($data);
    return $data;
  }
  elseif ($model['#type'] == 'user') {
    return (array) user_load($model_instance_id);
  }
  elseif ($model['#type'] == 'file') {
    return db_fetch_array(db_query("SELECT * FROM {files} WHERE fid=%d", $model_instance_id));
  }
  elseif ($model['#type'] == 'comment') {
    return array(_comment_load($model_instance_id));
  }
}


/**
 * Implementation of hook_ext_save_model_data.
 */
function ext_ext_save_model_data(&$data, $model_name, $new = FALSE) {
  $model = ext_get_models($model_name);

  if ($model['#type'] == 'view') {
    // NOTE: this is not implemented yet, see below TODO.
    return FALSE;
  }

  if ($model['#type'] == 'node') {
    // Prepare and convert data for node_save.
    if ($new) {
      unset($data['nid']);
      unset($data['vid']);
      if (!$data['uid']) {
        global $user;
        $data['uid'] = $user->uid;
      }
    }
    $data['type'] = $model['#node_type']; // Force correct node type.

    _ext_convert_to_cck_fields($data);

    // Do this after _ext_convert_to_cck_fields() as _ext_copy_filefield_files() expects regular CCK field format.
    _ext_copy_filefield_files($data, $new);

    $data = (object) $data; // node_save expects an object.
    $changed_orig = isset($data->changed) ? $data->changed : NULL;

    node_save($data);

    // Convert data back to our format.
    $data = (array) $data;
    _ext_convert_from_cck_fields($data);

    // If the node is new but didn't save.
    if (!$data['nid']) {
      return FALSE;
    }
    $changed_new = db_result(db_query("SELECT changed FROM {node} WHERE nid=%d", $data['nid']));
    // If the node wasn't new and didn't save.
    if ($changed_orig == $changed_new) {
      return FALSE;
    }
    return TRUE;
  }
  elseif ($model['#type'] == 'user') {
    if ($new) {
      unset($data['uid']);
    }
    // user_save expects an object.
    $data = (object) $data;
    $success = user_save($data);
    $data = (array) $data;
    return $success;
  }
  elseif ($model['#type'] == 'file') {
    // TODO: Should this be supported somehow?
    return FALSE;
  }
  elseif ($model['#type'] == 'comment') {
    // Unideally (see comments in ext_ext_access_model_data()) we have to check
    // for permission to post the comment to the specified node here.
    // This is done in the comment_save function.
    if ($new) {
      unset($data['cid']);
    }
    return comment_save($data);
  }
}


/**
 * Implementation of hook_ext_validate_model_data.
 */
function ext_ext_validate_model_data($data, $model_name) {
  $model = ext_get_models($model_name);
  if ($model['#type'] == 'node') {
    // node_validate expects an object.
    $data = (object) $data;
    _ext_convert_from_cck_fields($data);
    node_validate($data);
    return _ext_extract_form_errors();
  }
  elseif ($model['#type'] == 'comment') {
    comment_validate($data);
    return _ext_extract_form_errors();
  }
  return array();
}


/**
 * Implementation of hook_ext_delete_model_data.
 */
function ext_ext_delete_model_data($model_instance_id, $model_name) {
  $model = ext_get_models($model_name);
  if ($model['#type'] == 'node') {
    node_delete($model_instance_id);
    return db_result(db_query("SELECT COUNT(*) FROM {node} WHERE nid=%d", $model_instance_id)) == 0;
  }
  elseif ($model['#type'] == 'user') {
    user_delete(array(), $model_instance_id);
    return db_result(db_query("SELECT COUNT(*) FROM {users} WHERE uid=%d", $model_instance_id)) == 0;
  }
  elseif ($model['#type'] == 'comment') {
    // Below taken from comment_confirm_delete_submit().
    // Delete comment and its replies.
    $comment = _comment_load($model_instance_id);
    _comment_delete_thread($comment);
    _comment_update_node_statistics($comment->nid);
    // Clear the cache so an anonymous user sees that his comment was deleted.
    cache_clear_all();
  }
}


/**
 * Implementation of hook_ext_access_store_data.
 */
function ext_ext_access_store_data($store_name, $options, $account) {
  $store = ext_get_stores($store_name);

  if ($store['#type'] == 'node') {
    // Individual nodes are checked for access permissions in ext_ext_load_store_data().
    return user_access('access content');
  }
  if ($store['#type'] == 'user') {
    return user_access('administer users');
  }
  elseif ($store['#type'] == 'view') {
    list($model, $view_name, $display_name) = _ext_unpack_view_model($store_name);
    if ($view_name) {
      $view = views_get_view($view_name);
      if ($view) {
        return $view->access($display_name, $account);
      }
    }
  }
  return FALSE;
}


/**
 * Implementation of hook_ext_load_store_data.
 */
function ext_ext_load_store_data($store_name, $options) {
  $store = ext_get_stores($store_name);
  $limit = isset($options['limit']) ? $options['limit'] : 0;
  $start = isset($options['start']) ? $options['start'] : 0;

  if ($store['#type'] == 'node' || $store['#type'] == 'file' || $store['#type'] == 'comment') {
    $node_store = $store['#type'] == 'node';

    $model = ext_get_models($store['model']);
    $base_table = $model['#db_base_table'];
    $join_tables_type = array();
    $join_tables_field = array();
    $conditions = array();
    $arguments = array();
    $sorts = array();

    // If we're loading data for a node store for a specific node type.
    if ($node_store && $store['model'] != 'Node') {
      // Filter on node type for this Model.
      $conditions[] = "n.type = '%s'";
      $arguments[] = $model['#node_type'];
    }

    if (isset($options['filter'])) {
      // Turn the conditions into a query.
      foreach ($options['filter'] as $filter) {
        $field_def = $model['fields'][$filter['property']];
        $table = $field_def['#db_table'];
        $column = $field_def['#db_column'];

        if ($table != $base_table) {
          // Currently we only have joins for nodes (node_revisions and CCK tables).
          $join_tables_type[$table] = $table == 'node_revisions' ? 'LEFT JOIN' : 'INNER JOIN';
          $join_tables_field[$table] = 'vid';
        }

        // Some of below is taken from http://www.levihackwith.com/extjs-4-parsing-filters-from-the-filterfeature-using-php/
        // TODO: Some of it doesn't seem to apply and should be checked out
        // (eg the 'comparison', 'anyMatch' and 'caseSensitive' properties).
        if (isset($filter['comparison'])) {
          if ($filter['comparison'] == 'lt')
            $comparator = '<';
          elseif ($filter['comparison'] == 'gt')
            $comparator = '>';
          else
            $comparator = '=';
        }
        else {
          $comparator = '=';
        }

        $value = $filter['value'];
        $table_column = "{{$table}}.$column";
        $field_def['type'] = isset($field_def['type']) ? $field_def['type'] : 'string';
        switch($field_def['type']) {
          case 'boolean':
            $place_holder = "%b";
            break;
          case 'int':
            $place_holder = "%d";
            break;
          case 'float':
            $place_holder = "%f";
            break;
          case 'date':
            $place_holder = "'%s'";
            break;
          case 'string':
          default: // Assume string/date
            //'anyMatch' not passed through?
            //if (!empty($filter['anyMatch'])) {
              $place_holder = "'%%%s%%'";
              // Perhaps we should be using RegExs here, Ext seems to assume a regex for strings/filters?
              $comparator = 'LIKE';
            //}
            //else {
            //  $place_holder = "'%s'";
            //}

            // Assume case sensitive unless explicitly set otherwise.
            //'caseSensitive' not passed through?
            //if (isset($filter['caseSensitive']) && !$filter['caseSensitive']) {
            //  $table_column = "UPPER($table_column)";
            //  $place_holder = "UPPER($place_holder)";
            //}
            break;
        }
        $conditions[] = "$table_column $comparator $place_holder";
        $arguments[] = $value;
      }
    }

    if (isset($options['sort'])) {
      // Turn the conditions into a query.
      foreach ($options['sort'] as $sort) {
        $field_def = $model['fields'][$sort['property']];
        $direction = isset($sort['direction']) ? $sort['direction'] : 'ASC';

        $table = $field_def['#db_table'];
        $column = $field_def['#db_column'];
        if ($table != $base_table) {
          $join_tables_type[$table] = $table == 'node_revisions' ? 'LEFT JOIN' : 'INNER JOIN';
        }

        $sorts[] = "{{$table}}.$column $direction";
      }
    }

    $primary_id = $model['idProperty'];
    $base_table_alias = substr($base_table, 0, 1);
    $query = "SELECT $base_table_alias.$primary_id AS $primary_id FROM {$base_table} $base_table_alias";
    foreach ($join_tables_type as $table => $join) {
      $field = $join_tables_field[$table];
      $query .= " $join {{$table}} ON $base_table_alias.$field = {{$table}}.$field";
    }
    if (!empty($conditions)) {
      $query .= " WHERE ". implode(" AND ", $conditions);
    }
    if (!empty($sorts)) {
      $query .= " ORDER BY ". implode(', ', $sorts);
    }

    // Add access filters (and perhaps other stuff).
    $query = db_rewrite_sql($query, $base_table_alias, $primary_id);

    if ($limit > 0) {
      $results = db_query_range($query, $arguments, $start, $limit);
    }
    else {
      $results = db_query($query, $arguments);
    }
    $ids = array();
    while ($data = db_fetch_object($results)) {
      $ids[] = $data->$primary_id;
    }
    return $ids;
  }

  elseif ($store['#type'] == 'user') {
    // TODO: Add user query code. Have to deal with stupid profile fields.
  }

  elseif ($store['#type'] == 'view') {
    // Determine object type of the view.
    list($model, $view_name, $display_name) = _ext_unpack_view_model($store_name);

    // Unset these params so there's no chance of interfering with the View
    // arguments (these params are automatically produced by Ext).
    foreach (array('q', '_dc', 'limit', 'start', 'page') as $key) {
      unset($options[$key]);
    }

    $view = views_get_view($view_name);
    $view->set_display($display_name);
    $view->init_display();
    $arguments = array_keys($view->display_handler->get_option('arguments'));
    $args = array();
    $filters = array();
    if (isset($options['filter'])) {
      foreach ($options['filter'] as $filter) {
        $filters[$filter['property']] = $filter['value'];
      }
    }
    foreach ($arguments as $key) {
      $value = 'all';
      if (isset($filters[$key])) {
        $value = is_array($filters[$key]) ? implode('+', $filters[$key]) : $filters[$key];
      }
      $args[] = $value;
    }
    $view->set_arguments($args);
    $view->pre_execute();
    $view->pager['items_per_page'] = $limit;
    $view->pager['offset'] = $start;
    $view->pager['use_pager'] = FALSE;
    $view->execute();

    $return = array();
    $data = array();
    foreach ($view->result as $row) {
      $data[] = (array) $row;
    }
    unset($view); // Conserve some memory.

    // If a View based Model is being used (when display row style is 'fields').
    if ($model['#type'] == 'view') {
      // Just use raw result data from the View.
      $return = $data;
    }
    else {
      // Get the ID of the base object/table type the View.
      foreach ($data as $row) {
        $return[] = $row[$store['#primaryID']];
      }
    }
  }

  return $return;
}


/**
 * Given Model instance data for a 'node' Model type, convert from the
 * "separate Model field for each column of each CCK field" format to the
 * normal CCK field format as used in a node object (ie from
 * "data[field name_column name][delta] = value" format to
 * "data[field name][delta][column name] = value" format).
 * @param $data The Model instance data.
 */
function _ext_convert_to_cck_fields(&$data, $process_filefields = FALSE) {
  if (module_exists('content')) {
    // Get info about all CCK fields for all content types.
    $cck_fields = _content_type_info();
    // Construct CCK field value columns from separate fields.
    foreach ($cck_fields['content types'][$data['type']]['fields'] as $field_name => $field) {
      $data[$field_name] = array();
      $columns = array_keys($field['columns']);

      // Special handling to include extra columns/fields from files table for filefield fields.
      if ($field['type'] == 'filefield') {
        $file_fields = drupal_get_schema('files');
        $file_fields = $file_fields['fields'];
        unset($file_fields['fid']); //'fid' included in CCK filefield def.
        $columns = array_merge($columns, array_keys($file_fields));
      }

      foreach ($columns as $col) {
        // Get values as array.
        $values = $field['multiple'] ? $data[$field_name. '_'. $col] : array($data[$field_name. '_'. $col]);
        // $values can still be not an array in the case that the field allows
        // multiple values, but the Model instance field is empty or not set
        // (eg because the node is new).
        if (is_array($values)) {
          foreach ($values as $delta => $value) {
            $data[$field_name][$delta][$col] = $value;
          }
        }
        unset($data[$field_name. '_'. $col]);
      }
    }
  }
}


/**
 * Given a node object as an array, convert to the
 * "separate Model field for each column of each CCK field" format from the
 * normal CCK field format as used in a node object (ie to
 * "data[field name_column name][delta] = value" format from
 * "data[field name][delta][column name] = value" format).
 */
function _ext_convert_from_cck_fields(&$data) {
  if (module_exists('content')) {
    // Get info about all CCK fields for all content types.
    $cck_fields = _content_type_info();
    // Extract CCK field value columns into separate fields.
    foreach ($cck_fields['content types'][$data['type']]['fields'] as $field_name => $field) {
      $columns = array_keys($field['columns']);

      // Special handling to include extra columns/fields from files table for filefield fields.
      if ($field['type'] == 'filefield') {
        $file_fields = drupal_get_schema('files');
        $file_fields = $file_fields['fields'];
        unset($file_fields['fid']); //'fid' included in CCK filefield def.
        $columns = array_merge($columns, array_keys($file_fields));
      }

      $field_data_set = isset($data[$field_name]) && is_array($data[$field_name]);
      foreach ($columns as $col) {
        if ($field_data_set) {
          $values = _ext_extract_values_from_sub_arrays_by_key($col, $data[$field_name]);
          if ($field['multiple']) {
            $data[$field_name. '_'. $col] = $values;
          }
          else {
            if (count($values)) {
              $data[$field_name. '_'. $col] = $values[0];
            }
            else {
              $data[$field_name. '_'. $col] = NULL;
            }
          }
        }
        else {
          $data[$field_name. '_'. $col] = NULL;
        }
      }

      unset($data[$field_name]);
    }
  }
}


/**
 * Given a node containing filefield CCK fields, if a filefield references
 * existing files, then copy the actual files and create new records in the
 * 'files' table for them, and update the details in the filefield.
 * This is useful for a new node which has been created by copying an existing
 * node.
 *
 * @param $node The node. This is expected to be an array rather than the
 *   usual object.
 */
function _ext_copy_filefield_files(&$node, $new) {
  if (module_exists('content')) {
    global $user;

    // Get info about all CCK fields for all content types.
    $cck_fields = _content_type_info();
    foreach ($cck_fields['content types'][$node['type']]['fields'] as $field_name => $field) {
      if ($field['type'] == 'filefield' && isset($node[$field_name]) && is_array($node[$field_name])) {
        // For each file referenced by this filefield.
        foreach($node[$field_name] as $delta => &$data) {
          // If this is a new node or new file, and the file exists.
          if (($new || !$data['fid']) && $data['filepath'] && file_exists($data['filepath'])) {
            // Temporarily copy file to default files directory, filefield module will move it to correct location.
            $tmp_file_name = $data['filename']. '.'. mt_rand();
            $tmp_file_path = file_directory_path(). '/'. $tmp_file_name;

            if (copy($data['filepath'], $tmp_file_path)) {
              // Create new files record.
              $file = new stdClass();
              $file->filename = $tmp_file_name;
              $file->filepath = $tmp_file_path;

              $fileinfomime = finfo_open(FILEINFO_MIME_TYPE); // Return mime type ala mimetype extension.
              $file->filemime = !empty($data['filemime']) ? $data['filemime'] : finfo_file($fileinfomime, $tmp_file_path);
              $file->filesize = !empty($data['filesize']) ? $data['filesize'] : filesize($tmp_file_path);

              $file->uid = $user->uid;
              $file->status = FILE_STATUS_PERMANENT;
              $file->timestamp = time();

              // Save file to DB.
              drupal_write_record('files', $file);

              //Update filefield data.
              $data['fid'] = $file->fid;
              $data['filename'] = $file->filename;
              $data['filepath'] = $file->filepath;
              if (!isset($data['data'])) {
                $data['data'] = array();
              }
              $data['data']['process'] = TRUE; // Make filefield process the file.
            }
            else {
              watchdog('ext', "An error occurred while trying to copy a file from %source to %dest to copy a filefield." , array('%source' => $data['filepath'], '%dest' => $tmp_file_path), WATCHDOG_ERROR);
            }
          }
        }
      }
    }
  }
}


function _ext_gen_name($type, $part1 = NULL, $part2 = NULL) {
  switch ($type) {
    case 'node':
      return 'Node'. _ext_camelcase($part1);
    case 'view':
      return 'View'. _ext_camelcase($part1. 'VDN_'. $part2);
  }
}


function _ext_unpack_view_model($model_name, $instance_data = NULL) {
  list($view_name, $display_name) = explode('VDN', substr($model_name, 4));
  $view_name = _ext_uncamelcase($view_name);
  $display_name = _ext_uncamelcase($display_name);
  // See if a "custom" Model was defined for this view and display.
  $model = ext_get_models($model_name);
  if (!$model) {
    // Determine "standard" Model from view.
    $view = views_get_view($view_name);
    $model_name = _ext_base_table_to_model_name($view->base_table);
    if ($model_name == 'Node' && $instance_data != NULL) {
      // Determine content type specific Model name.
      $model_name = _ext_gen_name('node', $instance_data->type);
    }
    $model = ext_get_models($model_name);
  }
  return array($model, $view_name, $display_name);
}


function _ext_get_node_model($nid) {
  $name = 'Node';
  $node = node_load($nid);
  if ($node) {
    $name = _ext_gen_name('node', $node->type);
  }
  return ext_get_models($name);
}


function _ext_base_table_to_model_name($table) {
  static $table_to_model = array(
    'node' => 'Node',
    'users' => 'User',
    'comments' => 'Comment',
    'files' => 'File',
    'term_data' => 'Term',
  );
  return $table_to_model[$table];
}

// Determine primary ID field from schema for a table.
function _ext_table_primary_id($table) {
  $schema = drupal_get_schema($table);
  foreach ($schema['fields'] as $field_name => $field) {
    if ($field['type'] == 'serial') {
      return $field_name;
    }
  }
  return FALSE;
}

function _ext_db_schema_field_to_ext_field($field_name, $field, $db_table=NULL) {
  $ext_field = array(
    'name' => $field_name,
    'type' => _ext_db_schema_field_type_to_ext_field_type($field['type']),
  );
  if (isset($field['default'])) {
    $ext_field['defaultValue'] = $field['default'];
  }
  if ($db_table != NULL) {
    $ext_field['#db_table'] = $db_table;
    $ext_field['#db_column'] = $field_name;
  }
  return $ext_field;
}

function _ext_db_schema_field_type_to_ext_field_type($type) {
  static $type_map = array(
    'serial' => 'int',
    'int' => 'int',
    'float' => 'float',
    'numeric' => 'float',
    'varchar' => 'string',
    'char' => 'string',
    'text' => 'string',
    'blob' => 'string',
    'datetime' => 'date',
  );
  return $type_map[$type];
}


/**
 * This is used to extract errors set by form validation functions.
 * This is useful when validating model data using such functions.
 */
function _ext_extract_form_errors() {
  foreach (form_get_errors() as $key => $error) {
    $field = array_pop(explode('][', $key));
    $errors[$field] = $error; // Should already be translated.
  }
  form_set_error(NULL, NULL, TRUE); // Clear form errors.
  drupal_get_messages(); // Clear message queue.
  return $errors;
}